{********************************************************************}
{                                                                    }
{ written by TMS Software                                            }
{            Copyright (c) 2018 - 2019                               }
{            Email : info@tmssoftware.com                            }
{            Web : http://www.tmssoftware.com                        }
{                                                                    }
{ The source code is given as is. The author is not responsible      }
{ for any possible damage done due to the use of this code.          }
{ The complete source code remains property of the author and may    }
{ not be distributed, published, given or sold in any form as such.  }
{ No parts of the source code can be included in any other component }
{ or application without written authorization of the author.        }
{********************************************************************}

unit WEBLib.Lists;

interface

uses
  Classes, Web, WEBLib.Controls;

type
  TListItem = class;
  TListItems = class;
  TListControl = class;

  TListStyle = (lsNone, lsPagination, lsTabs, lsBreadcrumb, lsListGroup, lsListGroupHorizontal, lsNavPills, lsNavPillsHorizontal);

  TListGetItemClassEvent = procedure(Sender: TObject; AItem: TListItem; var AClassName: string) of object;

  TListGetItemChildrenEvent = procedure(Sender: TObject; AItem: TListItem; AElement: TJSHTMLElementRecord) of object;

  TListItemEvent = procedure(Sender: TObject; AListItem: TListItem) of object;


  TListItem = class(TCollectionItem)
  private
    FItems: TListItems;
    FLink: string;
    FText: string;
    FItemClassName: TElementClassName;
    FLinkClassName: TElementClassName;
    FTag: integer;
    FEnabled: boolean;
    FActive: boolean;
    FAutoCollaps: boolean;
    FListElement: TJSElement;
    FLinkTarget: string;
    FHint: string;
    procedure SetItems(const Value: TListItems);
    procedure SetActive(const Value: boolean);
    procedure SetEnabled(const Value: boolean);
    procedure SetText(const Value: string);
  protected
    function GetDisplayName: string; override;
    procedure Changed;
  public
    constructor Create(ACollection: TCollection); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    property ListElement: TJSElement read FListElement;
    procedure AddLink(const AURL: string; ATarget: string = ''); overload;
    procedure AddLink(const AURL: string; ATarget: string; AClass: string); overload;
    procedure Expand;
    procedure Collapse;
    function IsLink: boolean;
    function IsCollapsed: boolean;
  published
    property Active: boolean read FActive write SetActive default false;
    property AutoCollaps: boolean read FAutoCollaps write FAutoCollaps default false;
    property Enabled: boolean read FEnabled write SetEnabled default true;
    property Hint: string read FHint write FHint;
    property ItemClassName: TElementClassName read FItemClassName write FItemClassName;
    property Items: TListItems read FItems write SetItems;
    property Link: string read FLink write FLink;
    property LinkClassName: TElementClassName read FLinkClassName write FLinkClassName;
    property LinkTarget: string read FLinkTarget write FLinkTarget;
    property Tag: integer read FTag write FTag default 0;
    property Text: string read FText write SetText;
  end;

  TListItems = class(TOwnedCollection)
  private
    FOnChange: TNotifyEvent;
    function GetItem(Index: integer): TListItem;
    procedure SetItem(Index: integer; const Value: TListItem);
    function GetListControl: TListControl;
  protected
    procedure Update(Item: TCollectionItem); override;
  public
    constructor Create(AOwner: TPersistent);
    function Add: TListItem;
    function Insert(Index: integer): TListItem;
    property Items[Index: integer]: TListItem read GetItem write SetItem; default;
    property ListControl: TListControl read GetListControl;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

  TListControl = class(TWebCustomControl)
  private
    FItems: TListItems;
    FWidthStyle: TSizeStyle;
    FHeightPercent: TPercentSize;
    FHeightStyle: TSizeStyle;
    FWidthPercent: TPercentSize;
    FElementListClassName: TElementClassName;
    FOnGetItemChildren: TListGetItemChildrenEvent;
    FOnGetItemClass: TListGetItemClassEvent;
    FOnItemDblClick: TListItemEvent;
    FDefaultItemClassName: TElementClassName;
    FOnItemClick: TListItemEvent;
    FDefaultItemLinkClassName: TElementClassName;
    FListStyle: TListStyle;
    FItemIndex: integer;
    procedure SetItems(const Value: TListItems);
    procedure SetHeightPercent(const Value: TPercentSize);
    procedure SetHeightStyle(const Value: TSizeStyle);
    procedure SetWidthPercent(const Value: TPercentSize);
    procedure SetWidthStyle(const Value: TSizeStyle);
    procedure SetListStyle(const Value: TListStyle);
    procedure SetDefaultItemClassName(const Value: TElementClassName);
    procedure SetDefaultItemLinkClassName(const Value: TElementClassName);
  protected
    procedure ItemsChanged(Sender: TObject);
    procedure GetItemClass(Item: TListItem; var AClass: string); virtual;
    procedure GetItemChildren(Item: TListItem; AElement: TJSHTMLElement); virtual;
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property ItemIndex: integer read FItemIndex write FItemIndex;
  published
    property Align;
    property Anchors;
    property ChildOrder;
    property DefaultItemClassName: TElementClassName read FDefaultItemClassName write SetDefaultItemClassName;
    property DefaultItemLinkClassName: TElementClassName read FDefaultItemLinkClassName write SetDefaultItemLinkClassName;
    property DragMode;
    property ElementFont;
    property ElementPosition;
    property ElementListClassName: TElementClassName read FElementListClassName write FElementListClassName;
    property Height;
    property HeightStyle: TSizeStyle read FHeightStyle write SetHeightStyle default ssAbsolute;
    property HeightPercent: TPercentSize read FHeightPercent write SetHeightPercent;
    property Items: TListItems read FItems write SetItems;
    property Style: TListStyle read FListStyle write SetListStyle default lsNone;
    property Visible;
    property Width;
    property WidthStyle: TSizeStyle read FWidthStyle write SetWidthStyle default ssAbsolute;
    property WidthPercent: TPercentSize read FWidthPercent write SetWidthPercent;
    property OnGetItemChildren: TListGetItemChildrenEvent read FOnGetItemChildren write FOnGetItemChildren;
    property OnGetItemClass: TListGetItemClassEvent read FOnGetItemClass write FOnGetItemClass;
    property OnItemClick: TListItemEvent read FOnItemClick write FOnItemClick;
    property OnItemDblClick: TListItemEvent read FOnItemDblClick write FOnItemDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnStartDrag;
  end;

  {$IFNDEF LCLTMSWEB}
  [ComponentPlatforms(TMSWebPlatform)]
  {$ENDIF}
  TWebListControl = class(TListControl);

implementation

uses
  SysUtils, Graphics, WEBLib.HTML;

{ TListItems }

function TListItems.Add: TListItem;
begin
  Result := TListItem(inherited Add);
end;

constructor TListItems.Create(AOwner: TPersistent);
begin
  inherited Create(AOwner, TListItem);
end;

function TListItems.GetItem(Index: integer): TListItem;
begin
  Result := TListItem(inherited Items[Index]);
end;

function TListItems.GetListControl: TListControl;

  function FindListControl(AOwner: TPersistent): TListControl;
  begin
    Result := nil;

    if AOwner is TListControl then
      Result := AOwner as TListControl
    else
    begin
      if Assigned(AOwner) and (AOwner is TListItem) then
      begin
        Result := FindListControl((AOwner as TListItem).Collection.Owner);
      end;
    end;
  end;

begin
  Result := FindListControl(Self.Owner);
end;

function TListItems.Insert(Index: integer): TListItem;
begin
  Result := TListItem(inherited Insert(Index));
end;

procedure TListItems.SetItem(Index: integer; const Value: TListItem);
begin
  inherited Items[Index] := Value;
end;


procedure TListItems.Update(Item: TCollectionItem);
begin
  inherited;
  if Assigned(OnChange) then
    OnChange(Self);
end;

{ TListItem }

procedure TListItem.AddLink(const AURL: string; ATarget: string);
begin
  //
end;

procedure TListItem.AddLink(const AURL: string; ATarget, AClass: string);
begin
  //
end;

procedure TListItem.Assign(Source: TPersistent);
begin
  if (Source is TListItem) then
  begin
    FItemClassName := (Source as TListItem).ItemClassName;
    FText := (Source as TListItem).Text;
    FLink := (Source as TListItem).Link;
    FLinkClassName := (Source as TListItem).LinkClassName;
    FLinkTarget := (Source as TListItem).LinkTarget;
    FItems.Assign((Source as TListItem).Items);
    FTag := (Source as TListItem).Tag;
    FActive := (Source as TListItem).Active;
    FEnabled := (Source as TListItem).Enabled;
    FAutoCollaps := (Source as TListItem).AutoCollaps;
    FHint := (Source as TListItem).Hint;
  end;
end;

procedure TListItem.Changed;
begin
  Collection.BeginUpdate;
  Collection.EndUpdate;
end;

procedure TListItem.Collapse;
begin
//
end;

constructor TListItem.Create(ACollection: TCollection);
var
  ctrl: TListControl;

begin
  inherited Create(ACollection);
  FItems := TListItems.Create(Self);
  FActive := false;
  FEnabled := true;
  FAutoCollaps := false;
  FTag := 0;

  ctrl := (ACollection as TListItems).ListControl;

  if Assigned(ctrl) then
  begin
    FItemClassName := ctrl.DefaultItemClassName;
    FLinkClassName := ctrl.DefaultItemLinkClassName;
  end;
end;

destructor TListItem.Destroy;
begin
  FItems.Free;
  inherited Destroy;
end;

procedure TListItem.Expand;
begin
//
end;

function TListItem.GetDisplayName: string;
begin
  if Text <> '' then
    Result := Text
  else
    Result := 'Item '+ IntToStr(Index);
end;

function TListItem.IsCollapsed: boolean;
begin
  Result := true;
end;

function TListItem.IsLink: boolean;
begin
  Result := (Link <> '') or (LinkClassName <> '');
end;

procedure TListItem.SetActive(const Value: boolean);
begin
  FActive := Value;
  Changed;
end;

procedure TListItem.SetEnabled(const Value: boolean);
begin
  FEnabled := Value;
  Changed;
end;

procedure TListItem.SetItems(const Value: TListItems);
begin
  FItems.Assign(Value);
end;

procedure TListItem.SetText(const Value: string);
begin
  FText := Value;
  Changed;
end;

{ TListControl }

constructor TListControl.Create(AOwner: TComponent);
var
  FDesignTime: boolean;
begin
  inherited;
  FItems := TListItems.Create(Self);
  FItems.OnChange := ItemsChanged;

  FWidthStyle := ssAbsolute;
  FHeightStyle := ssAbsolute;
  FWidthPercent := 100;
  FHeightPercent := 100;
  ElementPosition := epAbsolute;
  Width := 300;
  Height := 60;

  FDesignTime := (csDesigning in ComponentState) and not
                 ((csReading in AOwner.ComponentState) or (csLoading in AOwner.ComponentState));

  if FDesignTime then
  begin
    FItems.Add.Text := 'Item 0';
    FItems.Add.Text := 'Item 1';
    FItems.Add.Text := 'Item 2';
  end;

end;

destructor TListControl.Destroy;
begin
  FItems.Free;
  inherited;
end;

procedure TListControl.GetItemChildren(Item: TListItem;
  AElement: TJSHTMLElement);
var
  LElementRec: TJSHTMLElementRecord;
begin
  if Assigned(OnGetItemChildren) then
  begin
    LElementRec.element := AElement;
    OnGetItemChildren(Self, Item, LElementRec);
  end;
end;

procedure TListControl.GetItemClass(Item: TListItem; var AClass: string);
begin
  if Assigned(OnGetItemClass) then
    OnGetItemClass(Self, Item, AClass);
end;

procedure TListControl.ItemsChanged(Sender: TObject);
begin
  Invalidate;
end;

procedure TListControl.Paint;
var
  i, th, x, y: integer;
  s: string;
begin

  Canvas.Font.Assign(Font);
  th := Canvas.TextHeight('gh') + 2;

  x := 2;
  y := 2;

  for i := 0 to Items.Count - 1 do
  begin
    s := HTMLStrip(Items[i].Text);
    if Trim(s) = ''  then
      s := 'Item ' + IntToStr(i);

    if Items[i].Active then
    begin
      Canvas.Brush.Style := bsSolid;
      Canvas.Brush.Color := clHighlight;
      Canvas.Font.Color := clHighlightText;
    end
    else
    begin
      Canvas.Brush.Style := bsClear;

      if not Items[i].Enabled then
        Canvas.Font.Color := clGray
      else
        Canvas.Font.Color := clWindowText;
    end;

    Canvas.TextOut(x,y, s);

    if (Style in [lsListGroupHorizontal, lsNavPillsHorizontal]) then
      x := x + Canvas.TextWidth(s) + 8
    else
      y := y + th;
  end;
end;

procedure TListControl.SetDefaultItemClassName(const Value: TElementClassName);
var
  i: integer;
begin
  FDefaultItemClassName := Value;

  if (csDesigning in ComponentState) and not (csLoading in ComponentState) and (Items.Count > 0) then
  begin
    for i := 0 to Items.Count - 1  do
    begin
      Items[i].ItemClassName := Value;
    end;
  end;
end;

procedure TListControl.SetDefaultItemLinkClassName(
  const Value: TElementClassName);
var
  i: integer;
begin
  FDefaultItemLinkClassName := Value;
  if (csDesigning in ComponentState) and not (csLoading in ComponentState) and (Items.Count > 0) then
  begin
    for i := 0 to Items.Count - 1  do
    begin
      Items[i].LinkClassName := Value;
    end;
  end;
end;

procedure TListControl.SetHeightPercent(const Value: TPercentSize);
begin
  FHeightPercent := Value;
end;

procedure TListControl.SetHeightStyle(const Value: TSizeStyle);
begin
  FHeightStyle := Value;
end;

procedure TListControl.SetItems(const Value: TListItems);
begin
  FItems.Assign(Value);
end;

procedure TListControl.SetListStyle(const Value: TListStyle);
var
  FDesignTime: boolean;

begin
  FListStyle := Value;

  FDesignTime := (csDesigning in ComponentState) and not
                 ({(csReading in Owner.ComponentState) or} (csLoading in Owner.ComponentState));

  if FDesignTime then
  begin

    case FListStyle of
      lsPagination:
        begin
          DefaultItemClassName := 'page-item';
          DefaultItemLinkClassName := 'page-link';
          ElementListClassName := 'pagination';
        end;
      lsTabs:
        begin
          DefaultItemClassName := 'nav-item';
          DefaultItemLinkClassName := 'nav-link';
          ElementListClassName := 'nav nav-tabs';
        end;
      lsBreadcrumb:
        begin
          DefaultItemClassName := 'breadcrumb-item';
          DefaultItemLinkClassName := 'breadcrumb-link';
          ElementListClassName := 'breadcrumb';
        end;
      lsListGroup:
        begin
          DefaultItemClassName := 'list-group-item';
          DefaultItemLinkClassName := 'list-group-link';
          ElementListClassName := 'list-group';
        end;
      lsListGroupHorizontal:
        begin
          DefaultItemClassName := 'list-group-item';
          DefaultItemLinkClassName := 'list-group-link';
          ElementListClassName := 'list-group list-group-horizontal';
        end;
      lsNavPills:
        begin
          DefaultItemClassName := 'nav-item';
          DefaultItemLinkClassName := 'nav-link';
          ElementListClassName := 'nav nav-pills flex-column';
        end;
      lsNavPillsHorizontal:
        begin
          DefaultItemClassName := 'nav-item';
          DefaultItemLinkClassName := 'nav-link';
          ElementListClassName := 'nav nav-pills';
        end;
    end;
  end;

  Invalidate;
end;

procedure TListControl.SetWidthPercent(const Value: TPercentSize);
begin
  FWidthPercent := Value;
end;

procedure TListControl.SetWidthStyle(const Value: TSizeStyle);
begin
  FWidthStyle := Value;
end;

end.
