{********************************************************************}
{                                                                    }
{ written by TMS Software                                            }
{            Copyright (c) 2018 - 2020                               }
{            Email : info@tmssoftware.com                            }
{            Web : http://www.tmssoftware.com                        }
{                                                                    }
{ The source code is given as is. The author is not responsible      }
{ for any possible damage done due to the use of this code.          }
{ The complete source code remains property of the author and may    }
{ not be distributed, published, given or sold in any form as such.  }
{ No parts of the source code can be included in any other component }
{ or application without written authorization of the author.        }
{********************************************************************}

unit WEBLib.SQLCDS;

interface

uses
  SysUtils, Classes, WEBLib.Controls, DB
  {$IFNDEF LCLTMSWEB}
  , DBClient
  {$ENDIF}
  , WEBLib.CDS;

type
  ERestError = class(EDatabaseError);

  // Stub
  TDatarequest = class(TObject)
  end;

  TRecordUpdateDescriptor = class(TObject)
  end;

  TRestGetURLEvent = procedure (Sender : TComponent; aRequest : TDataRequest; var aURL : string) of object;
  TRestUpdateURLEvent = procedure (Sender : TComponent; aRequest : TRecordUpdateDescriptor; var aURL : string) of object;

  {$IFNDEF LCLTMSWEB}
  [ComponentPlatforms(TMSWebPlatform)]
  {$ENDIF}
  TSQLRestConnection = class(TWebClientConnection)
  private
    FonGetResources: TNotifyEvent;
    FLastError: string;
    FMetaDataResourceName : string;
    FConnectionsResourceName: string;
    FCustomViewResourceName: string;
    FDatabaseConnections: TStrings;
    FResources: TStrings;
    { Private declarations }
  protected
    // These are only used at design time for caching purposes, don't use in code !
    procedure ClearError;
    property LastError : String Read FLastError Write FLastError;
    property DesignDatabaseConnections : TStrings Read FDatabaseConnections;
    property DesignResources : TStrings Read FResources;
  public
    { Public declarations }
    constructor Create(aOwner : TComponent); override;
    destructor Destroy; override;
  published
    { Published declarations }
    property OnGetResources : TNotifyEvent read FonGetResources write FOnGetResources;
    property MetaDataResourceName : string read FMetaDataResourceName write FMetaDataResourceName;
    property ConnectionsResourceName : string read FConnectionsResourceName write FConnectionsResourceName;
    property CustomViewResourceName : string read FCustomViewResourceName write FCustomViewResourceName;
  end;

  {$IFNDEF LCLTMSWEB}
  [ComponentPlatforms(TMSWebPlatform)]
  {$ENDIF}
  TWebSQLRestConnection = Class(TSQLRestConnection);

  {$IFNDEF LCLTMSWEB}
  [ComponentPlatforms(TMSWebPlatform)]
  {$ENDIF}
  TWebSQLRestClientDataset = class(TWebClientDataSet)
  private
    FResourceName: string;
    FConnection: TSQLRestConnection;
    FDatabaseConnection: string;
    FSQL: TStrings;
    FParams: TParams;
    FUseServerMetaData: boolean;
    procedure SetConnection(const Value: TSQLRestConnection);
    procedure SetSQL(const Value: TStrings);
    procedure DoSQLChange(Sender: TObject);
    procedure SetResourceName(const Value: string);
    procedure SetParams(const Value: TParams);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(aOwner : TComponent); override;
    destructor Destroy; override;
    function MyURL : string;
    function CustomViewResourceName: string;
    function CleanSQL: string;
  published
    property Params: TParams read FParams write SetParams;
    property Connection: TSQLRestConnection read FConnection write SetConnection;
    property DatabaseConnection: string read FDatabaseConnection write FDatabaseConnection;
    property ResourceName: string read FResourceName write SetResourceName;
    property SQL: TStrings read FSQL write SetSQL;
    property IDField;
    property UseServerMetaData: boolean read FUseServerMetaData write FUseServerMetaData default false;
    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
    property BeforeRefresh;
    property AfterRefresh;
    property OnCalcFields;
    property OnDeleteError;
    property OnEditError;
    property OnFilterRecord;
    property OnNewRecord;
    property OnPostError;
  end;


implementation

{$IFNDEF LCLTMSWEB}
uses
  System.NetEncoding;
{$ENDIF}

{$IFDEF LCLTMSWEB}
function URLEncode(const Url: string): string;
var
  i: Integer;
  UrlA: ansistring;
  res: ansistring;

begin
  res := '';
  UrlA := ansistring(UTF8Encode(Url));

  for i := 1 to Length(UrlA) do
  begin
    case UrlA[i] of
      'A'..'Z', 'a'..'z', '0'..'9', '-', '_', '.':
        res := res + UrlA[i];
    else
        res := res + '%' + ansistring(IntToHex(Ord(UrlA[i]), 2));
    end;
  end;

  Result := string(res);
end;
{$ENDIF}

{ TWebSQLRestClientDataset }

constructor TWebSQLRestClientDataset.Create(aOwner: TComponent);
begin
  inherited;
  FSQL := TStringList.Create;
  TStringList(FSQL).OnChange := DoSQLChange;
  FParams := TParams.Create(TParam);
end;

destructor TWebSQLRestClientDataset.Destroy;
begin
  FreeAndNil(FParams);
  TStringList(FSQL).OnChange:=Nil;
  FreeAndNil(FSQL);
  inherited;
end;

procedure TWebSQLRestClientDataset.DoSQLChange(Sender : TObject);

begin
  if Trim(FSQL.Text) <> '' then
    FResourceName := CustomViewResourceName;
end;

function TWebSQLRestClientDataset.CustomViewResourceName : String;

begin
  if Assigned(Connection) then
    Result := Connection.CustomViewResourceName
  else
    Result := 'customView';
end;

function TWebSQLRestClientDataset.CleanSQL: String;

begin
  Result := StringReplace(SQL.Text,#13#10,' ',[rfReplaceAll]);
  Result := StringReplace(Result,#10,' ',[rfReplaceAll]);
  Result := StringReplace(Result,#10,' ',[rfReplaceAll]);
end;

function TWebSQLRestClientDataset.MyURL: String;

begin
  Result := DatabaseConnection;
  if (Result <> '') and (Result[Length(Result)] <> '/') then
    Result := Result+'/';
  Result := Result + ResourceName;
  if SameText(ResourceName,CustomViewResourceName) then
    Result := Result + '?SQL=' + CleanSQL;

  {$IFNDEF LCLTMSWEB}
  Result := TNetEncoding.URL.Encode(Result);
  {$ENDIF}
  {$IFDEF LCLTMSWEB}
  Result := URLEncode(Result);
  {$ENDIF}
end;

procedure TWebSQLRestClientDataset.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) and (AComponent = FConnection) then
    FConnection := nil;
end;

procedure TWebSQLRestClientDataset.SetConnection(const Value: TSQLRestConnection);
begin
  if (FConnection = Value) then
    Exit;

  If Assigned(FCOnnection) then
    FConnection.RemoveFreeNotification(Self);

  FConnection := Value;

  If Assigned(FCOnnection) then
    FConnection.FreeNotification(Self);
end;

procedure TWebSQLRestClientDataset.SetParams(const Value: TParams);
begin
  if FParams = Value then
    Exit;
  FParams.Assign(Value);
end;

procedure TWebSQLRestClientDataset.SetResourceName(const Value: String);
begin
  if Value = FResourceName then
    Exit;
  if Not SameText(Value,CustomViewResourceName) then
    FSQL.Clear;
  FResourceName := Value;
end;

procedure TWebSQLRestClientDataset.SetSQL(const Value: TStrings);
begin
  FSQL.Assign(Value);
end;

{ TSQLRestConnection }

procedure TSQLRestConnection.ClearError;
begin
  FLastError := '';
end;

constructor TSQLRestConnection.Create(aOwner: TComponent);
begin
  inherited;
  FDatabaseConnections := TStringList.Create;
  FResources := TStringList.Create;
  ConnectionsResourceName := '_connection';
  MetaDataResourceName := 'metadata';
  CustomViewResourceName := 'customview';
end;

destructor TSQLRestConnection.Destroy;

begin
  FreeAndNil(FDatabaseConnections);
  FreeAndNil(FResources);
  inherited;
end;

end.
